#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cuda_runtime.h>

void print_to_file(int* arr, int arr_len, char* type) {
    char filename[100];
    time_t now = time(NULL);
    struct tm *t = localtime(&now);
    strftime(filename, sizeof(filename) - 1, "output/%Y_%m_%d_%H%M%S", t);
    snprintf(filename, sizeof(filename) - 1, "%s-%s.txt", filename, type);
    FILE *f = fopen(filename, "w");
    for (int i = 0; i < arr_len; i++) {
        fprintf(f, "%d ", arr[i]);
    }
    fclose(f);
}

void rng(int* arr, int n) {
    int seed = 13516095;
    srand(seed);
    for(long i = 0; i < n; i++) {
        arr[i] = (int)rand();
    }
}

int* down_sweep(int* flags, int n) {
    int* prefix_sum = (int*) malloc(n * sizeof(int));
    prefix_sum[0] = 0;

    for (int i = 1; i < n; i++) {
        prefix_sum[i] = prefix_sum[i - 1] + flags[i - 1];
    }

    return prefix_sum;
}

int* up_sweep(int* flags, int n) {
    int* prefix_sum = (int*) malloc(n * sizeof(int));
    prefix_sum[n - 1] = n - 1;

    for (int i = n - 2; i >= 0; i--) {
        int substract = (flags[i + 1] ? 0 : 1);
        prefix_sum[i] = prefix_sum[i + 1] - substract;
    }

    return prefix_sum;
}

__global__ void assign_index(int* arr, int* arr_idx, int* down_sweep_i, int* up_sweep_i, int* flags, int n) {
    for (int i = blockIdx.x * blockDim.x + threadIdx.x; i < n; i += blockDim.x * gridDim.x) {
        arr_idx[i] = flags[i] ? down_sweep_i[i] : up_sweep_i[i];
    }
}

__global__ void assign_flags(int* arr, int n, int idx, int* flags) {
    for (int i = blockIdx.x * blockDim.x + threadIdx.x; i < n; i += blockDim.x * gridDim.x) {
        flags[i] = ((arr[i] & 1 << idx) == 1 << idx) ? 0 : 1;
    }
}

__global__ void scatter(int* arr, int* arr_temp, int* arr_idx, int n) {
    for (int i = blockIdx.x * blockDim.x + threadIdx.x; i < n; i += blockDim.x * gridDim.x) {
        arr[arr_idx[i]] = arr_temp[i];
    }
}

void count_sort(int* d_arr, int n, int idx) {
    int block_dim = 64;
    int grid_dim = (n + block_dim - 1) / block_dim;
    int*  d_flags;
    cudaMalloc(&d_flags, n * sizeof(int));
    assign_flags<<<grid_dim, block_dim>>>(d_arr, n, idx, d_flags);
    cudaDeviceSynchronize();
    int* flags = (int*) malloc(n * sizeof(int));
    cudaMemcpy(flags, d_flags, n * sizeof(int), cudaMemcpyDeviceToHost);
    int* down_sweep_i = down_sweep(flags, n);
    int* up_sweep_i = up_sweep(flags, n);
    free(flags);
    int *d_arr_temp, *d_arr_idx, *d_idx_down, *d_idx_up;
    cudaMalloc(&d_arr_temp, n * sizeof(int));
    cudaMalloc(&d_arr_idx, n * sizeof(int));
    cudaMalloc(&d_idx_down, n * sizeof(int));
    cudaMalloc(&d_idx_up, n * sizeof(int));
    cudaMemcpy(d_idx_down, down_sweep_i, n * sizeof(int), cudaMemcpyHostToDevice);
    cudaMemcpy(d_idx_up, up_sweep_i, n * sizeof(int), cudaMemcpyHostToDevice);
    cudaMemcpy(d_arr_temp, d_arr, n * sizeof(int), cudaMemcpyDeviceToDevice);
    assign_index<<<grid_dim, block_dim>>>(d_arr, d_arr_idx, d_idx_down, d_idx_up, d_flags, n);
    cudaDeviceSynchronize();
    scatter<<<grid_dim, block_dim>>>(d_arr, d_arr_temp, d_arr_idx, n);
    cudaDeviceSynchronize();
    cudaFree(d_arr_idx);
    cudaFree(d_idx_down);
    cudaFree(d_idx_up);
    cudaFree(d_flags);
    cudaFree(d_arr_temp);
}

void radix_sort(int* d_arr, int n) { 
    for (int idx = 1; idx < 32; idx++) {
        count_sort(d_arr, n, idx);
    }
}

void print(int* arr, int n) { 
    for (int i = 0; i < n; i++) 
        printf("%d ", arr[i]);
    printf("\n");
} 

int main(int argc, char *argv[]) {
    if (argc != 2) {
        printf("Usage: ./radix_sort <n>\n");
        return 1;
    }

    // Generate numbers
    int n = atoi(argv[1]);
    int* arr = (int*) malloc(sizeof(int) * n);
    rng(arr, n);
    
    // Sort numbers
    clock_t start, end;
    double cpu_time_used;

    printf("Sorting in parallel...\n");
    start = clock();

    // Move sorting CPU
    int* d_arr;
    cudaMalloc(&d_arr, sizeof(int) * n);
    cudaMemcpy(d_arr, arr, sizeof(int) * n, cudaMemcpyHostToDevice);
    radix_sort(d_arr, n);
    cudaMemcpy(arr, d_arr, n * sizeof(int), cudaMemcpyDeviceToHost);
    cudaFree(d_arr);

    end = clock();
    cpu_time_used = ((double) (end - start)) * 1000000 / CLOCKS_PER_SEC;
    printf("Parallel radix sort finished, time taken: %f μs\n\n", cpu_time_used);
    print_to_file(arr, n, "output-parallel");

    return 0; 
} 
